@@ -1,14 +1,12 @@
 import os
-import time
 import json
 import math
-from typing import List
+from typing import List, Tuple
 import platform
 import logging
 import os, json, shutil
 from glob import glob
 import concurrent.futures
-from datetime import datetime
 
 from seg2map import exceptions
 from seg2map import common
@@ -20,10 +18,131 @@ import tqdm
 import tqdm.auto
 import tqdm.asyncio
 import ee
+import zipfile
+from area import area
+import geopandas as gpd
+from shapely.geometry import LineString, MultiPolygon, Polygon
+from shapely.ops import split
+from osgeo import gdal
 
 
 logger = logging.getLogger(__name__)
 
+import asyncio
+import aiohttp
+import logging
+from aiohttp import ClientResponseError, ClientConnectionError
+
+
+async def async_download_url_dict(url_dict: dict = {}):
+    async with aiohttp.ClientSession() as session:
+        tasks = []
+        for save_path, url in url_dict.items():
+            task = asyncio.create_task(
+                download_data_with_retries(session, url, save_path, total_attempts=2)
+            )
+            tasks.append(task)
+        results = await tqdm.asyncio.tqdm.gather(*tasks)
+
+
+async def download_data_with_retries(
+    session,
+    url,
+    save_location,
+    retries=3,
+    initial_delay=1,
+    max_delay=32,
+    total_attempts=2,
+):
+    attempts = 0
+    while attempts < total_attempts:
+        success = await download_with_retries(
+            session, url, save_location, retries, initial_delay, max_delay
+        )
+        if success:
+            return True  # Download successful
+        attempts += 1
+
+    logger.error(f"Download failed after {total_attempts} attempts: {url}")
+    print(f"Download failed after {total_attempts} attempts: {url}")
+    return False
+
+
+async def async_download_url(session, url: str, save_path: str):
+    model_name = url.split("/")[-1]
+    chunk_size: int = 2048
+    async with session.get(url, raise_for_status=True) as r:
+        content_length = r.headers.get("Content-Length")
+        if content_length is not None:
+            content_length = int(content_length)
+            with open(save_path, "wb") as fd:
+                with tqdm.auto.tqdm(
+                    total=content_length,
+                    unit="B",
+                    unit_scale=True,
+                    unit_divisor=1024,
+                    desc=f"Downloading {model_name}",
+                    initial=0,
+                    ascii=False,
+                    position=0,
+                ) as pbar:
+                    async for chunk in r.content.iter_chunked(chunk_size):
+                        fd.write(chunk)
+                        pbar.update(len(chunk))
+        else:
+            with open(save_path, "wb") as fd:
+                async for chunk in r.content.iter_chunked(chunk_size):
+                    fd.write(chunk)
+
+
+async def download_with_retries(
+    session, url, save_location, retries, initial_delay, max_delay
+):
+    delay = initial_delay
+    for i in range(retries):
+        try:
+            async with session.get(url) as response:
+                response.raise_for_status()
+                content_length = response.headers.get("Content-Length")
+                if content_length is not None:
+                    content_length = int(content_length)
+                    with open(save_location, "wb") as fd:
+                        with tqdm.auto.tqdm(
+                            total=content_length,
+                            unit="B",
+                            unit_scale=True,
+                            unit_divisor=1024,
+                            desc=f"Downloading {os.path.dirname(save_location)}",
+                            initial=0,
+                            ascii=False,
+                            position=0,
+                        ) as pbar:
+                            async for chunk in response.content.iter_chunked(1024):
+                                if not chunk:
+                                    break
+                                fd.write(chunk)
+                                pbar.update(len(chunk))
+                else:
+                    with open(save_location, "wb") as fd:
+                        async for chunk in response.content.iter_chunked(1024):
+                            fd.write(chunk)
+                return True  # Download successful
+        except (ClientResponseError, ClientConnectionError) as e:
+            logger.error(f"Error downloading {url}: {e}")
+        except asyncio.exceptions.TimeoutError as e:
+            logger.error(f"Timeout error for {url}: {e}")
+        except Exception as e:
+            logger.error(f"Unexpected error downloading {url}: {e}")
+        if i < retries - 1:
+            logger.warning(
+                f"Retrying download in {delay} seconds... ({i + 1}/{retries})"
+            )
+            await asyncio.sleep(delay)
+            # Update delay for the next iteration, doubling it while ensuring it doesn't exceed max_delay
+            delay = min(delay * 2, max_delay)
+        else:
+            return False
+
 
 async def download_file(session, url, save_location):
     retries = 3  # number of times to retry download
@@ -47,6 +166,9 @@ async def download_file(session, url, save_location):
             print(
                 f"Timeout error occurred for {url}. Retrying with new session in 1 second... ({i + 1}/{retries})"
             )
+            logger.warning(
+                f"Timeout error occurred for {url}. Retrying with new session in 1 second... ({i + 1}/{retries})"
+            )
             await asyncio.sleep(1)
             async with aiohttp.ClientSession() as new_session:
                 return await download_file(new_session, url, save_location)
@@ -55,6 +177,9 @@ async def download_file(session, url, save_location):
             logger.error(
                 f"Download failed for {save_location} {url}. Retrying in 1 second... ({i + 1}/{retries})"
             )
+            logger.warning(
+                f"Timeout error occurred for {url}. Retrying with new session in 1 second... ({i + 1}/{retries})"
+            )
             print(
                 f"Download failed for {url}. Retrying in 1 second... ({i + 1}/{retries})"
             )
@@ -73,7 +198,6 @@ async def download_group(session, group, semaphore):
         filepath = os.path.abspath(tile["filepath"])
         filenames = {
             "multiband": "multiband" + str(tile_number),
-            "singleband": os.path.basename(filepath),
         }
         for tile_id in tile["ids"]:
             logger.info(f"tile_id: {tile_id}")
@@ -103,7 +227,7 @@ async def download_group(session, group, semaphore):
     common.unzip_dir(group[0]["filepath"])
     common.delete_empty_dirs(group[0]["filepath"])
     # delete duplicate tifs. keep tif with most non-black pixels
-    common.delete_tifs_at_same_location(group[0]["filepath"])
+    # common.delete_tifs_at_same_location(group[0]["filepath"])
 
 
 # Download the information for each year
@@ -213,38 +337,6 @@ def run_async_function(async_callback, **kwargs) -> None:
     return result
 
 
-import json
-import math
-import logging
-import os, json, shutil
-from glob import glob
-import concurrent.futures
-from datetime import datetime
-import platform
-
-from seg2map import exceptions
-from seg2map import common
-
-from typing import List, Tuple
-
-import tqdm
-import tqdm.auto
-import zipfile
-from area import area
-import numpy as np
-import geopandas as gpd
-import asyncio
-import aiohttp
-import tqdm.asyncio
-import nest_asyncio
-from shapely.geometry import LineString, MultiPolygon, Polygon
-from shapely.ops import split
-import ee
-from osgeo import gdal
-
-logger = logging.getLogger(__name__)
-
-
 # GEE allows for 20 concurrent requests at once
 limit = asyncio.Semaphore(20)
 
@@ -413,10 +505,7 @@ def create_tasks(
     polygon: List[tuple],
     tile_id: str,
     filepath: str,
-    multiband_filepath: str,
-    filenames: dict,
-    file_id: str,
-    download_bands: str,
+    filename: str,
 ) -> list:
     """
 
@@ -434,17 +523,10 @@ def create_tasks(
         ex: 'USDA/NAIP/DOQQ/m_4012407_se_10_1_20100612'
     filepath : str
         full path to tile directory that data will be saved to
-    multiband_filepath : str
-        The path to a directory will save multiband files
-    filenames : dict
-        A dictionary of filenames:
-        'singleband': name of singleband files
-        'multiband': name of multiband files
+    filename : str
+       name of file that data will be downloaded to
     file_id : str
         name that file will be saved as based on tile_id
-    download_bands : str
-        type of imagery to download
-        must be one of the following strings "multiband","singleband", or "both"
 
     Returns
     -------
@@ -453,35 +535,21 @@ def create_tasks(
 
     """
     tasks = []
-    if download_bands == "multiband" or download_bands == "both":
-
-        task = asyncio.create_task(
-            async_download_tile(
-                session,
-                polygon,
-                tile_id,
-                multiband_filepath,
-                filename=filenames["multiband"] + "_" + file_id,
-                filePerBand=False,
-            )
+    task = asyncio.create_task(
+        async_download_tile(
+            session,
+            polygon,
+            tile_id,
+            filepath,
+            filename,
+            filePerBand=False,
         )
-        tasks.append(task)
-    if download_bands == "singleband" or download_bands == "both":
-        task = asyncio.create_task(
-            async_download_tile(
-                session,
-                polygon,
-                tile_id,
-                filepath,
-                filename=filenames["singleband"] + "_" + file_id,
-                filePerBand=False,
-            )
-        )
-        tasks.append(task)
+    )
+    tasks.append(task)
     return tasks
 
 
-async def async_download_all_tiles(tiles_info: List[dict], download_bands: str) -> None:
+async def async_download_all_tiles(tiles_info: List[dict]) -> None:
     # creates task for each tile to be downloaded and waits for tasks to complete
     tasks = []
     for counter, tile_dict in enumerate(tiles_info):
@@ -489,13 +557,7 @@ async def async_download_all_tiles(tiles_info: List[dict], download_bands: str)
         filepath = os.path.abspath(tile_dict["filepath"])
         parent_dir = os.path.dirname(filepath)
         multiband_filepath = os.path.join(parent_dir, "multiband")
-        filenames = {
-            "multiband": "multiband" + str(counter),
-            "singleband": os.path.basename(filepath),
-        }
-
-        timeout = aiohttp.ClientTimeout(total=3000)
-        async with aiohttp.ClientSession(timeout=timeout) as session:
+        async with aiohttp.ClientSession(timeout=3000) as session:
             for tile_id in tile_dict["ids"]:
                 logger.info(f"tile_id: {tile_id}")
                 file_id = tile_id.replace("/", "_")
@@ -503,6 +565,8 @@ async def async_download_all_tiles(tiles_info: List[dict], download_bands: str)
                 year_str = file_id.split("_")[-1][:4]
                 if len(file_id.split("_")[-2]) == 8:
                     year_str = file_id.split("_")[-2][:4]
+
+                filename = "multiband" + str(counter) + "_" + file_id
                 # full path to year directory within multiband dir eg. ./multiband/2012
                 year_filepath = os.path.join(multiband_filepath, year_str)
                 logger.info(f"year_filepath: {year_filepath}")
@@ -513,9 +577,8 @@ async def async_download_all_tiles(tiles_info: List[dict], download_bands: str)
                         tile_id,
                         filepath,
                         year_filepath,
-                        filenames,
+                        filename,
                         file_id,
-                        download_bands,
                     )
                 )
     # show a progress bar of all the requests in progress
@@ -725,8 +788,6 @@ async def get_tiles_for_ids(
         roi_gdf (gpd.GeoDataFrame): geodataframe of ROIs on the map
         ids (List[str]): ids of ROIs to download imagery for
         dates (Tuple[str]): start and end dates
-        download_bands (str): type of imagery to download
-            must be one of the following strings "multiband","singleband", or "both"
     """
     ROI_tiles = {}
     tasks = []
@@ -753,14 +814,12 @@ async def get_tiles_for_ids(
     return ROI_tiles
 
 
-async def async_download_ROIs(ROI_tiles: List[dict], download_bands: str) -> None:
+async def async_download_ROIs(ROI_tiles: List[dict]) -> None:
     """
     Downloads the specified bands for each ROI tile asynchronously using aiohttp and asyncio.
 
     Parameters:
     ROI_tiles (List[dict]): A list of dictionaries representing the ROI tiles to download.
-    download_bands (str): A comma-separated string of band numbers to download.
-
     Returns:
     None: This function does not return anything.
     """
@@ -770,23 +829,18 @@ async def async_download_ROIs(ROI_tiles: List[dict], download_bands: str) -> Non
         for ROI_tile in ROI_tiles:
             for year in ROI_tile.keys():
                 print(f"YEAR for ROI tile: {year}")
-                task = asyncio.create_task(
-                    async_download_year(ROI_tile[year], download_bands, session)
-                )
+                task = asyncio.create_task(async_download_year(ROI_tile[year], session))
                 tasks.append(task)
         # show a progress bar of all the requests in progress
         await tqdm.asyncio.tqdm.gather(*tasks, position=0, desc=f"All Downloads")
 
 
-async def async_download_year(
-    tiles_info: List[dict], download_bands: str, session
-) -> None:
+async def async_download_year(tiles_info: List[dict], session) -> None:
     """
     Downloads the specified bands for each tile in the specified year asynchronously using aiohttp and asyncio.
 
     Parameters:
     tiles_info (List[dict]): A list of dictionaries representing the tiles to download.
-    download_bands (str): A comma-separated string of band numbers to download.
     session: The aiohttp session to use for downloading.
 
     Returns:
@@ -800,7 +854,6 @@ async def async_download_year(
         filepath = os.path.abspath(tile_dict["filepath"])
         filenames = {
             "multiband": "multiband" + str(counter),
-            "singleband": os.path.basename(filepath),
         }
         for tile_id in tile_dict["ids"]:
             logger.info(f"tile_id: {tile_id}")
@@ -815,7 +868,6 @@ async def async_download_year(
                     filepath,
                     filenames,
                     file_id,
-                    download_bands,
                 )
             )
     # show a progress bar of all the requests in progress
@@ -823,14 +875,3 @@ async def async_download_year(
     common.unzip_data(os.path.dirname(filepath))
     # delete any directories that were empty
     common.delete_empty_dirs(os.path.dirname(filepath))
-
-
-# call asyncio to run download_ROIs
-def run_magic_function_to_download(ROI_tiles: List[dict], download_bands: str) -> None:
-    if platform.system() == "Windows":
-        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
-    # apply a nested loop to jupyter's event loop for async downloading
-    nest_asyncio.apply()
-    # get nested running loop and wait for async downloads to complete
-    loop = asyncio.get_running_loop()
-    loop.run_until_complete(async_download_ROIs(ROI_tiles, download_bands))
