@@ -1,6 +1,5 @@
 # standard python imports
 import os
-import glob
 import logging
 
 # internal python imports
@@ -63,22 +62,6 @@ class UI_Models:
     # all instances of UI will share the same debug_view
     model_view = Output()
     run_model_view = Output()
-    # uav_RGB_10class_7566797: https://zenodo.org/record/7566797
-    # uav_RGB_10class_7566810: https://zenodo.org/record/7566810
-    # CoastTrain
-    # ortho_RGB_8class_7574784: https://zenodo.org/record/7574784
-    # naip_RGB_5class_7566992: https://zenodo.org/record/7566992
-    # naip_RGB_8class_7570583: https://zenodo.org/record/7570583
-    # CHESAPEAKE
-    # ortho_RGB_7class_7576904:  https://zenodo.org/record/7576904
-    # OPEN EARTH NET
-    # ortho_RGB_9class_7576894: https://zenodo.org/record/7576894
-    # DeepGlobe
-    # ortho_RGB_7clas_7576898 https://zenodo.org/record/7576898
-    # EnviroAtlas
-    # ortho_RGB_6class_7576909 https://zenodo.org/record/7576909
-    # AAAI-Buildings
-    # ortho_RGB_7class_7607895 https://zenodo.org/record/7607895
 
     def __init__(self):
         # Controls size of ROIs generated on map
@@ -166,7 +149,7 @@ class UI_Models:
         model_choices_box = HBox(
             [self.model_type_dropdown, self.model_dropdown, self.model_implementation]
         )
-        checkboxes = HBox([self.GPU_checkbox, self.otsu_radio, self.tta_radio])
+        checkboxes = HBox([self.otsu_radio, self.tta_radio])
         instr_vbox = VBox(
             [
                 self.instr_select_images,
@@ -275,7 +258,7 @@ class UI_Models:
 
         self.instr_select_images = HTML(
             value="<b>1. Select Images</b> \
-                <br> - Select multiband directory within an ROI directory",
+                <br> - Select an ROI directory or a directory containing at least one ROI subdirectory.\nExample: ./data/dataset1/ID_e7CxBi_dates_2010-01-01_to_2014-12-31",
             layout=Layout(margin="0px 0px 0px 20px"),
         )
 
@@ -342,32 +325,36 @@ class UI_Models:
                 "Must click 'Select Images' first",
             )
             return
+        if not common.check_id_subdirectories_exist(inputs_directory):
+            self.launch_error_box(
+                "Cannot Run Model",
+                "You must select a directory that contains ROI subdirectories. Example ROI name: 'ID_e7CxBi_dates_2010-01-01_to_2014-12-31'",
+            )
+            return
         # Disable run and open results buttons while the model is running
-        # self.open_results_button.disabled = True
-        # self.run_model_button.disabled = True
+        self.run_model_button.disabled = True
 
         # gets GPU or CPU depending on whether use_GPU is True
         use_GPU = self.model_dict["use_GPU"]
         model_implementation = self.model_dict["implementation"]
-        model_name = self.model_dict["model_type"]
+        model_id = self.model_dict["model_type"]
         use_otsu = self.model_dict["otsu"]
         use_tta = self.model_dict["tta"]
 
-        zoo_model.get_GPU(use_GPU)
-        zoo_model_instance = zoo_model.Zoo_Model()
-
-        zoo_model_instance.run_model(
-            model_implementation,
-            session_name=session_name,
-            src_directory=inputs_directory,
-            model_name=model_name,
-            use_GPU=use_GPU,
-            use_otsu=use_otsu,
-            use_tta=use_tta,
+        zoo_model_instance = zoo_model.ZooModel()
+        try:
+            zoo_model_instance.run_model(
+                model_implementation,
+                session_name=session_name,
+                src_directory=inputs_directory,
+                model_id=model_id,
+                use_GPU=use_GPU,
+                use_otsu=use_otsu,
+                use_tta=use_tta,
         )
-        # # Enable run and open results buttons when model has executed
-        # self.run_model_button.disabled = False
-        # self.open_results_button.disabled = False
+        finally:
+            # Enable run and open results buttons when model has executed
+            self.run_model_button.disabled = False
 
     @run_model_view.capture(clear_output=True)
     def open_results_button_clicked(self, button):
